/*
** Nokia 5110 Fast & Furious Program
** Target: ATMEGA :: AVR-GCC
**
** Written by Ion Cristian-Razvan - 2016
*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <stdio.h>
#include <stdlib.h>
#include "lcd.h"

#define BUFF_X_MAX		(LCD_X_RES)
#define BUFF_Y_MAX		(LCD_Y_RES / 8)

#define BLOCK_SIZE		4
#define BORDER_LEFT		0
#define BORDER_RIGHT	49

#define NUM_BLOCKS		4

#define MAX_BLOCKS_X	((BORDER_RIGHT - 1) / 4)
#define MAX_BLOCKS_Y	((LCD_Y_RES - 1) / 4)

#define MAX_LEVEL		5

/* Define useful enums */
typedef enum {PANEL_START, PANEL_GAME, PANEL_OVER} PANEL;
typedef enum {BUTTON_LEFT = PC0, BUTTON_RIGHT = PC1} BUTTON;
typedef enum {PIECE_1, PIECE_2, PIECE_3, PIECE_4} PIECE;
typedef enum {DIRECTION_LEFT, DIRECTION_RIGHT} DIRECTION;
typedef enum {COLOR_BLACK, COLOR_WHITE} COLOR;
typedef enum {FALSE = 0, TRUE = 1} BOOL;

/* Define useful structs */
typedef struct {
	uint8_t x;
	uint8_t y;
} coordinate_t;

typedef struct {
	PIECE type;
	coordinate_t blcp[NUM_BLOCKS]; // bottom left corner pixels
} piece_t;

typedef struct  {
	PANEL panel;
	BOOL over;
	uint64_t entropy;
	uint8_t level;
	uint8_t score;
	piece_t pieces[10];
	piece_t car;
	uint8_t lcd_buff[BUFF_Y_MAX][BUFF_X_MAX];
} ff_game;

/* Function prototypes */
static void my_delay(uint8_t num);
static BOOL pressed_button(BUTTON btn);
static void set_pixel(uint8_t x, uint8_t y, COLOR color);
static void clear_screen(void);
static void block_draw(uint8_t xs, uint8_t ys);
static void block_erase(uint8_t xs, uint8_t ys);
static void current_piece_get(void);
static void current_piece_draw(void);
static void draw_car(void);
static void current_piece_erase(void);
static void car_erase(void);
static void current_piece_move(void);
static void car_move(DIRECTION dir);
static BOOL collision_with_placed_pieces_left(void);
static BOOL collision_with_placed_pieces_right(void);
static void draw_borders(void);
static void draw_start_panel(void);
static void draw_over_panel(void);
static void draw_score_panel(void);
static void update_score_panel(void);
static void treat_input(void);
static void init_game(void);
static uint32_t temper(uint32_t x);
static uint32_t lcg64_temper(uint64_t *seed);

/* Global variables */
ff_game game;
int nr;
int no_pieces;

int main(void)
{
	/* Setup LCD */
	lcd_init();
	lcd_contrast(0x40);

	/* Draw start panel and init game */
	draw_start_panel();
	init_game();
	my_delay(1);

	/* Switch through the game panel */
	while (TRUE) {
		switch (game.panel) {
		case PANEL_START:
			if (pressed_button(BUTTON_LEFT)) {
				draw_borders();
				draw_score_panel();
				draw_car();
				no_pieces++;
				current_piece_get();
				current_piece_draw();
				game.panel = PANEL_GAME;
			}
			break;
		case PANEL_GAME:
			if (game.over) {
				draw_over_panel();
				game.panel = PANEL_OVER;
				my_delay(1);
			} else {
				treat_input();
				for (uint8_t i = 0; i < game.level; i++) {
					current_piece_erase();
					current_piece_move();
					current_piece_draw();
				}
				my_delay(5);
			}
			break;
		case PANEL_OVER:
			if (pressed_button(BUTTON_LEFT)) {
			draw_start_panel();
			init_game();
			game.panel = PANEL_START;
			my_delay(1);
			}
			break;
		default: return 1;
		}
	}

	return 0;
}

/*
 * Stalls the program by 1000 * num.
 */
static void my_delay(uint8_t num)
{
	for (uint8_t i = 0; i < num; i++)
		_delay_ms(1000);
}

/*
 * Returns whether the button was pressed.
 */
static BOOL pressed_button(BUTTON btn)
{
	return (PINC & (1 << btn)) == 0 ? TRUE : FALSE;
}

/*
 * Sets the pixel at (x, y) coordinates to the requested color.
 */
static void set_pixel(uint8_t x, uint8_t y, COLOR color)
{
	uint8_t yy = LCD_Y_RES - 1 - y;
	uint8_t row = yy / 8;
	uint8_t val = game.lcd_buff[row][x];

	/* Compute pixel value based on color */
	switch (color) {
	case COLOR_BLACK:
		val |= (1 << (yy % 8));
		break;
	case COLOR_WHITE:
		val &= ~(1 << (yy % 8));
		break;
	default: return;
	}

	game.lcd_buff[row][x] = val;

	/* Move LCD pointer to (x, y) coordinates */
	lcd_send(0x80 | x, LCD_CMD);    // column
	lcd_send(0x40 | row, LCD_CMD);  // row

	/* Set the pixel */
	lcd_send(val, LCD_DATA);
}

/*
 * Clears the whole LCD screen (as in set all pixels to WHITE).
 */
static void clear_screen(void)
{
	for (uint8_t x = 0; x < LCD_X_RES; x++)
		for (uint8_t y = 0; y < LCD_Y_RES; y++)
			set_pixel(x, y, COLOR_WHITE);
}

/*
 * Draws a square block of BLOCK_SIZE size starting at bottom left corner defined by (xs, ys).
 */
static void block_draw(uint8_t xs, uint8_t ys)
{
	for (uint8_t x = xs; x < xs + BLOCK_SIZE; x++)
		for (uint8_t y = ys; y < ys + BLOCK_SIZE; y++)
			set_pixel(x, y, COLOR_BLACK);
}

/*
 * Erases a square block of BLOCK_SIZE size starting at bottom left corner defined by (xs, ys).
 */
static void block_erase(uint8_t xs, uint8_t ys)
{
	for (uint8_t x = xs; x < xs + BLOCK_SIZE; x++)
		for (uint8_t y = ys; y < ys + BLOCK_SIZE; y++)
			set_pixel(x, y, COLOR_WHITE);
}

/*
 * Gets a random piece type based on game's entropy and places it to the start position.
 */
static void current_piece_get(void)
{
	/* Sadly <time.h> was not available */
	game.pieces[no_pieces].type = lcg64_temper(&game.entropy) % 4;

	/*
	 *   *
	 *  ***
	 */
	switch (game.pieces[no_pieces].type) {
	case PIECE_1:
		game.pieces[no_pieces].blcp[0].x = 1;
		game.pieces[no_pieces].blcp[0].y = 37;
		game.pieces[no_pieces].blcp[1].x = game.pieces[no_pieces].blcp[0].x + 1 * BLOCK_SIZE;
		game.pieces[no_pieces].blcp[1].y = game.pieces[no_pieces].blcp[0].y;
		game.pieces[no_pieces].blcp[2].x = game.pieces[no_pieces].blcp[0].x + 2 * BLOCK_SIZE;
		game.pieces[no_pieces].blcp[2].y = game.pieces[no_pieces].blcp[0].y;
		game.pieces[no_pieces].blcp[3].x = game.pieces[no_pieces].blcp[0].x + 1 * BLOCK_SIZE;
		game.pieces[no_pieces].blcp[3].y = game.pieces[no_pieces].blcp[0].y + 1 * BLOCK_SIZE;
		break;
	case PIECE_2:
		game.pieces[no_pieces].blcp[0].x = 13;
		game.pieces[no_pieces].blcp[0].y = 37;
		game.pieces[no_pieces].blcp[1].x = game.pieces[no_pieces].blcp[0].x + 1 * BLOCK_SIZE;
		game.pieces[no_pieces].blcp[1].y = game.pieces[no_pieces].blcp[0].y;
		game.pieces[no_pieces].blcp[2].x = game.pieces[no_pieces].blcp[0].x + 2 * BLOCK_SIZE;
		game.pieces[no_pieces].blcp[2].y = game.pieces[no_pieces].blcp[0].y;
		game.pieces[no_pieces].blcp[3].x = game.pieces[no_pieces].blcp[0].x + 1 * BLOCK_SIZE;
		game.pieces[no_pieces].blcp[3].y = game.pieces[no_pieces].blcp[0].y + 1 * BLOCK_SIZE;
		break;
	case PIECE_3:
		game.pieces[no_pieces].blcp[0].x = 25;
		game.pieces[no_pieces].blcp[0].y = 37;
		game.pieces[no_pieces].blcp[1].x = game.pieces[no_pieces].blcp[0].x + 1 * BLOCK_SIZE;
		game.pieces[no_pieces].blcp[1].y = game.pieces[no_pieces].blcp[0].y;
		game.pieces[no_pieces].blcp[2].x = game.pieces[no_pieces].blcp[0].x + 2 * BLOCK_SIZE;
		game.pieces[no_pieces].blcp[2].y = game.pieces[no_pieces].blcp[0].y;
		game.pieces[no_pieces].blcp[3].x = game.pieces[no_pieces].blcp[0].x + 1 * BLOCK_SIZE;
		game.pieces[no_pieces].blcp[3].y = game.pieces[no_pieces].blcp[0].y + 1 * BLOCK_SIZE;
		break;
	case PIECE_4:
		game.pieces[no_pieces].blcp[0].x = 37;
		game.pieces[no_pieces].blcp[0].y = 37;
		game.pieces[no_pieces].blcp[1].x = game.pieces[no_pieces].blcp[0].x + 1 * BLOCK_SIZE;
		game.pieces[no_pieces].blcp[1].y = game.pieces[no_pieces].blcp[0].y;
		game.pieces[no_pieces].blcp[2].x = game.pieces[no_pieces].blcp[0].x + 2 * BLOCK_SIZE;
		game.pieces[no_pieces].blcp[2].y = game.pieces[no_pieces].blcp[0].y;
		game.pieces[no_pieces].blcp[3].x = game.pieces[no_pieces].blcp[0].x + 1 * BLOCK_SIZE;
		game.pieces[no_pieces].blcp[3].y = game.pieces[no_pieces].blcp[0].y + 1 * BLOCK_SIZE;
		break;
	default: return;
	}
}

/*
 * Draws the pieces block by block.
 */
static void current_piece_draw(void)
{
	if (game.pieces[0].blcp[NUM_BLOCKS - 1].y <= 1) {
		for (uint8_t j = 0; j <= no_pieces; j++) {
			game.pieces[j].type = game.pieces[j + 1].type;
			for (uint8_t i = 0; i < NUM_BLOCKS; i++) {
				game.pieces[j].blcp[i].x = game.pieces[j + 1].blcp[i].x;
				game.pieces[j].blcp[i].y = game.pieces[j + 1].blcp[i].y;
			}
		}
		no_pieces--;
	}

	for (uint8_t j = 0; j <= no_pieces; j++)
		for (uint8_t i = 0; i < NUM_BLOCKS; i++)
			if (game.pieces[j].blcp[i].y >= 1)
				block_draw(game.pieces[j].blcp[i].x, game.pieces[j].blcp[i].y);
}

/*
 * Draws the car.
 */
static void draw_car(void)
{
	for (uint8_t i = 0; i < NUM_BLOCKS; i++)
		block_draw(game.car.blcp[i].x, game.car.blcp[i].y);
}

/*
 * Erases the pieces block by block.
 */
static void current_piece_erase(void)
{
	for (uint8_t j = 0; j <= no_pieces; j++)
		for (uint8_t i = 0; i < NUM_BLOCKS; i++)
			if (game.pieces[j].blcp[i].y >= 1)
				block_erase(game.pieces[j].blcp[i].x, game.pieces[j].blcp[i].y);
}

/*
 * Erases the car.
 */
static void car_erase(void)
{
	for (uint8_t i = 0; i < NUM_BLOCKS; i++)
		block_erase(game.car.blcp[i].x, game.car.blcp[i].y);
}

/*
 * Moves the pieces one block down.
 */
static void current_piece_move()
{
	if (collision_with_placed_pieces_left() || collision_with_placed_pieces_right()) {
		game.over = TRUE;
		return;
	}

	for (uint8_t j = 0; j <= no_pieces; j++)
		for (uint8_t i = 0; i < NUM_BLOCKS; i++)
			if (game.pieces[j].blcp[i].y - BLOCK_SIZE >= 1)
				game.pieces[j].blcp[i].y -= BLOCK_SIZE;
			else
				game.pieces[j].blcp[i].y = 0;

	nr++;

	if (nr == 5) {
		game.score++;
		if (game.score % 10 == 0)
			game.level++;
		update_score_panel();

		no_pieces++;
		current_piece_get();
		nr = 0;
	}
}

/*
 * Moves the car one block in the specified direction.
 */
static void car_move(DIRECTION dir)
{
	switch (dir) {
	case DIRECTION_LEFT:
		/* Check if moving left is possible (does not collide with left border or another placed block) */
		for (uint8_t i = 0; i < NUM_BLOCKS; i++) {
			if (game.car.blcp[i].x == BORDER_LEFT + 1)
				return;
			if (collision_with_placed_pieces_left()) {
				game.over = TRUE;
				return;
			}
		}

		/* Move the piece one block to the left */
		for (uint8_t i = 0; i < NUM_BLOCKS; i++)
			game.car.blcp[i].x -= BLOCK_SIZE;
		break;
	case DIRECTION_RIGHT:
		/* Check if moving right is possible (does not collide with right border or another placed block) */
		for (uint8_t i = 0; i < NUM_BLOCKS; i++) {
			if (game.car.blcp[i].x + BLOCK_SIZE == BORDER_RIGHT)
				return;
			if (collision_with_placed_pieces_right()) {
				game.over = TRUE;
				return;
			}
		}

		/* Move the piece one block to the right */
		for (uint8_t i = 0; i < NUM_BLOCKS; i++)
			game.car.blcp[i].x += BLOCK_SIZE;
		break;
	default: return;
	}
}

/*
 * Checks whether the current piece collides with the already placed pieces on the left side.
 */
static BOOL collision_with_placed_pieces_left(void)
{
	for (uint8_t j = 0; j <= no_pieces; j++)
		for (uint8_t i = 0; i < NUM_BLOCKS; i++)
			for (uint8_t k = 0; k < NUM_BLOCKS; k++)
				if (game.pieces[j].blcp[i].x + BLOCK_SIZE > game.car.blcp[k].x &&
					game.pieces[j].blcp[i].x <= game.car.blcp[k].x &&
					game.pieces[j].blcp[i].y < game.car.blcp[k].y + BLOCK_SIZE &&
					game.pieces[j].blcp[i].y >= game.car.blcp[k].y)
					return TRUE;

	return FALSE;
}

/*
 * Checks whether the current piece collides with the already placed pieces on the right side.
 */
static BOOL collision_with_placed_pieces_right(void)
{
	for (uint8_t j = 0; j <= no_pieces; j++)
		for (uint8_t i = 0; i < NUM_BLOCKS; i++)
			for (uint8_t k = 0; k < NUM_BLOCKS; k++)
				if (game.pieces[j].blcp[i].x < game.car.blcp[k].x + BLOCK_SIZE &&
					game.pieces[j].blcp[i].x >= game.car.blcp[k].x &&
					game.pieces[j].blcp[i].y < game.car.blcp[k].y + BLOCK_SIZE &&
					game.pieces[j].blcp[i].y >= game.car.blcp[k].y)
					return TRUE;

	return FALSE;
}

/*
 * Draws the borders on the LCD.
 */
static void draw_borders(void)
{
	clear_screen();

	for (uint8_t y = 0; y < LCD_Y_RES; y++) {
		set_pixel(BORDER_LEFT, y, COLOR_BLACK);
		set_pixel(BORDER_RIGHT, y, COLOR_BLACK);
	}

	for (uint8_t x = 0; x < BORDER_RIGHT; x++)
		set_pixel(x, 0, COLOR_BLACK);
}

/*
 * Draws the text on the start panel.
 */
static void draw_start_panel(void)
{
	clear_screen();
	lcd_goto_xy(2, 1);
	lcd_str("Cristian Ion");
	lcd_goto_xy(6, 3);
	lcd_str("F & F");
	lcd_goto_xy(3, 5);
	lcd_str("Press LEFT");
	lcd_goto_xy(4, 6);
	lcd_str("to start.");
}

/*
 * Draws the text on the game over panel.
 */
static void draw_over_panel(void)
{
	char buf[11];
	sprintf(buf, "Score: %03u", game.score);
	clear_screen();
	lcd_goto_xy(3, 1);
	lcd_str("GAME OVER!");
	lcd_goto_xy(3, 3);
	lcd_str(buf);
	lcd_goto_xy(3, 5);
	lcd_str("Press LEFT");
	lcd_goto_xy(3, 6);
	lcd_str("to replay.");
}

/*
 * Draws the text on the score panel.
 */
static void draw_score_panel(void)
{
	lcd_goto_xy(10, 1);
	lcd_str("Score");
	lcd_goto_xy(11, 2);
	lcd_str("000");
	lcd_goto_xy(11, 5);
	lcd_str("Lvl");
	lcd_goto_xy(12, 6);
	lcd_str("1");
}

/*
 * Updates the score panel with the new score, lines and level.
 */
static void update_score_panel(void)
{
	char buf[4];
	lcd_goto_xy(11, 2);
	sprintf(buf, "%03u", game.score);
	lcd_str(buf);
	lcd_goto_xy(12, 6);
	lcd_chr('0' + game.level);
}

/*
 * Treats the input coming from the buttons.
 */
static void treat_input(void)
{
	if (pressed_button(BUTTON_LEFT)) {
		/* BUTTON_LEFT means move left */
		for (uint8_t i = 0; i < game.level; i++) {
			car_erase();
			car_move(DIRECTION_LEFT);
			draw_car();
		}
	} else if (pressed_button(BUTTON_RIGHT)) {
		/* BUTTON_RIGHT means move right */
		for (uint8_t i = 0; i < game.level; i++) {
			car_erase();
			car_move(DIRECTION_RIGHT);
			draw_car();
		}
	}
}

/*
 * Inits the game by resetting all its attributes.
 */
static void init_game(void)
{
	/* Set all port C pins as inputs and activate pull-up resitors */
	DDRC = 0x00;
	PORTC = 0xFF;

	/* Disable JTAG port */
	MCUCR = _BV(JTD);
	MCUCR = _BV(JTD);

	/* Reset game's attributes */
	game.panel = PANEL_START;
	game.over = FALSE;
	game.level = 1;
	game.score = 0;

	nr = 0;
	no_pieces = -1;

	game.car.blcp[0].x = 13;
	game.car.blcp[0].y = 1;
	game.car.blcp[1].x = game.car.blcp[0].x + 1 * BLOCK_SIZE;
	game.car.blcp[1].y = game.car.blcp[0].y;
	game.car.blcp[2].x = game.car.blcp[0].x + 2 * BLOCK_SIZE;
	game.car.blcp[2].y = game.car.blcp[0].y;
	game.car.blcp[3].x = game.car.blcp[0].x + 1 * BLOCK_SIZE;
	game.car.blcp[3].y = game.car.blcp[0].y + 1 * BLOCK_SIZE;
}

/*
 * https://stackoverflow.com/questions/19083566/what-are-the-better-pseudo-random-number-generator-than-the-lcg-for-lottery-sc
 */
static uint32_t temper(uint32_t x)
{
	x ^= x >> 11;
	x ^= x << 7 & 0x9D2C5680;
	x ^= x << 15 & 0xEFC60000;
	x ^= x >> 18;
	return x;
}

/*
 * https://stackoverflow.com/questions/19083566/what-are-the-better-pseudo-random-number-generator-than-the-lcg-for-lottery-sc
 */
static uint32_t lcg64_temper(uint64_t *seed)
{
	*seed = 6364136223846793005ULL * *seed + 1;
	return temper(*seed >> 32);
}
